@@ -2,7 +2,7 @@
  * This file is part of tiptop.
  *
  * Author: Erven ROHOU
- * Copyright (c) 2011, 2012 Inria
+ * Copyright (c) 2011, 2012, 2014, 2015 Inria
  *
  * License: GNU General Public License version 2.
  *
@@ -23,6 +23,7 @@
 #include "hash.h"
 #include "options.h"
 #include "pmc.h"
+#include "priv.h"
 #include "process.h"
 #include "screen.h"
 #include "spawn.h"
@@ -155,20 +156,92 @@ static char* get_cmdline(int pid, char* result, int size)
 }
 
 
+
+void start_counters(struct process* ptr,
+                    const screen_t* const screen,
+                    struct STRUCT_NAME* events,
+                    const struct option* const options)
+{
+  const int cpu = -1;
+  const int grp = -1;
+  const int flags = 0;
+  int zz;
+
+  /* Get number of counters from screen */
+  ptr->num_events = screen->num_counters;
+
+  /* If we have reached the maximum number of open files, we try to
+     close the events attached to some idle processes (unless
+     forbidden to do so by command line flag). */
+  if ((num_files + ptr->num_events >= num_files_limit) &&
+      (!options->no_collect))
+  {
+    int num_collected = 0;
+    struct process* q = ptr->next;
+    while (q && (num_collected < ptr->num_events)) {
+      if (q->cpu_percent < options->cpu_threshold) {
+        for(zz = 0; zz < q->num_events; zz++) {
+          if (q->fd[zz] >= 0) {
+            close(q->fd[zz]);
+            q->fd[zz] = -1;
+            q->values[zz] = 0xffffffff;
+            num_collected++;
+          }
+        }
+      }
+      q = q->next;
+    }
+    num_files -= num_collected;
+  }
+  
+  for(zz = 0; zz < ptr->num_events; zz++)
+    ptr->prev_values[zz] = 0;
+
+  /* restore super powers, if any, for the time of the system call */
+  restore_privilege();
+  for(zz = 0; zz < ptr->num_events; zz++) {
+    int fd;
+    events->type = screen->counters[zz].type;  /* eg PERF_TYPE_HARDWARE */
+    events->config = screen->counters[zz].config;
+
+    if (num_files < num_files_limit) {
+      fd = perf_event_open(events, ptr->tid, cpu, grp, flags);
+      if (fd == -1) {
+        error_printf("Could not attach counter '%s' to PID %d (%s): %s\n",
+                     screen->counters[zz].alias,
+                     ptr->tid,
+                     ptr->name,
+                     strerror(errno));
+      }
+    }
+    else {
+      fd = -1;
+      error_printf("Files limit reached for PID %d (%s)\n",
+                   ptr->tid, ptr->name);
+    }
+
+    if (fd != -1)
+      num_files++;
+    ptr->fd[zz] = fd;
+    ptr->values[zz] = 0;
+  }
+
+  /* drop super powers again */
+  drop_privilege();
+}
+
+
 void new_processes(struct process_list* const list,
                    const screen_t* const screen,
                    const struct option* const options)
 {
-  struct dirent* pid_dirent;
-  DIR*           pid_dir;
-  int            num_tids, val, n;
+  struct dirent*     pid_dirent;
+  DIR*               pid_dir;
+  int                num_tids, val, n, num_inactive, alloc_inact, i;
   struct STRUCT_NAME events = {0, };
-  FILE*          f;
-  uid_t          my_uid = -1;
-
-  const int cpu = -1;
-  const int grp = -1;
-  const int flags = 0;
+  FILE*              f;
+  uid_t              my_uid = -1;
+  struct process**   inactive;
 
   /* To avoid scanning the entire /proc directory, we first check if
      any process has been created since last time. /proc/loadavg
@@ -192,6 +265,10 @@ void new_processes(struct process_list* const list,
   if (options->show_kernel == 0)
     events.exclude_kernel = 1;
 
+  num_inactive = 0;
+  alloc_inact = 100;
+  inactive = malloc(alloc_inact * sizeof(struct process*));
+
   /* check all directories of /proc */
   pid_dir = opendir("/proc");
   while ((pid_dirent = readdir(pid_dir))) {
@@ -260,8 +337,8 @@ void new_processes(struct process_list* const list,
        root's processes because they are too many). */
     skip_by_user = 1;
     my_uid = options->euid;
-    if (((my_uid != 0) && (uid == my_uid)) ||  /* not root, monitor mine */
-        ((my_uid == 0) && (uid != 0)))         /* I am root, monitor all others */
+    if (((my_uid != 0) && (uid == my_uid)) || /* not root, monitor mine */
+        ((my_uid == 0) && (uid != 0)))        /* I am root, monitor all others*/
       skip_by_user = 0;
 
     if ((skip_by_user == 0) && (skip_by_pid == 0)) {
@@ -277,7 +354,6 @@ void new_processes(struct process_list* const list,
 
       /* Iterate over all threads in the process */
       while ((thr_dirent = readdir(thr_dir))) {
-        int   zz, fail;
         struct process* ptr;
         struct passwd*  passwd;
 
@@ -289,13 +365,15 @@ void new_processes(struct process_list* const list,
         if (ptr)  /* already known */
           continue;
 
-
         /* We have a new thread. */
 
         /* allocate memory */
         ptr = malloc(sizeof(struct process));
 
         /* insert into list of processes */
+        /* Insert at the front of the list. This will later let us
+           scan the rest of the list when looking for idle processes,
+           closing files. */
         ptr->next = list->processes;
         list->processes = ptr;
 
@@ -313,9 +391,6 @@ void new_processes(struct process_list* const list,
         ptr->pid = pid;
         ptr->proc_id = -1;
         ptr->dead = 0;
-#if 0
-        ptr->attention = 0;
-#endif
         ptr->u.d = 0.0;
 
         passwd = getpwuid(uid);
@@ -337,62 +412,58 @@ void new_processes(struct process_list* const list,
         ptr->cpu_percent_s = 0.0;
         ptr->cpu_percent_u = 0.0;
 
-        /* Get number of counters from screen */
-        ptr->num_events = screen->num_counters;
+        float uptime;
+        f = fopen("/proc/uptime", "r");
+        n = fscanf(f, "%f", &uptime);
+        fclose(f);
+        uptime *= clk_tck;
 
-        for(zz = 0; zz < ptr->num_events; zz++)
-          ptr->prev_values[zz] = 0;
+        /* read utime, stime and starttime (fields 14, 15, and 22) */
+        snprintf(name, sizeof(name) - 1, "/proc/%d/stat", tid);
+        f = fopen(name, "r");
+        if (!f)
+          continue;
+        long utime, stime;
+        long long starttime;
+        n = fscanf(f, "%*d (%*[^)]) %*c %*d %*d %*d %*d %*d %*u %*u "
+                   "%*u %*u %*u %lu %lu %*d %*d %*d %*d %*d %*d %llu",
+                   &utime, &stime, &starttime);
+        fclose(f);
+
+        /* Due to the limited number of files in a Linux process (each
+           counter corresponds to a file), we want to start counters
+           first for active process. Idle processes are
+           postponed. This increases the chances to display useful
+           information: a dash only for idle processes. */
+        if ((utime + stime)/(uptime - starttime) > 0.3) {
+          /* active process: %CPU > 30% */
+          start_counters(ptr, screen, &events, options);
+        }
+        else {
+          /* less active: postpone. Add to a list of inactive
+             processes, to be handled later. */
+          inactive[num_inactive++] = ptr;
+          if (num_inactive == alloc_inact) {
+            alloc_inact += 100;
+            inactive = realloc(inactive, alloc_inact * sizeof(struct process*));
+          }
+        }
 
         ptr->txt = malloc(TXT_LEN * sizeof(char));
 
-        fail = 0;
-        for(zz = 0; zz < ptr->num_events; zz++) {
-          int fd;
-          events.type = screen->counters[zz].type;  /* eg PERF_TYPE_HARDWARE */
-          events.config = screen->counters[zz].config;
-
-          if (num_files < num_files_limit) {
-            fd = sys_perf_counter_open(&events, tid, cpu, grp, flags);
-            if (fd == -1) {
-              error_printf("Could not attach counter '%s' to PID %d (%s): %s\n",
-                           screen->counters[zz].alias,
-                           tid,
-                           ptr->name,
-                           strerror(errno));
-            }
-          }
-          else {
-            fd = -1;
-            error_printf("Files limit reached for PID %d (%s)\n",
-                         tid, ptr->name);
-          }
-
-          if (fd == -1)
-            fail++;
-          else
-            num_files++;
-          ptr->fd[zz] = fd;
-          ptr->values[zz] = 0;
-        }
-
-#if 0
-        if (fail) {
-          /* at least one counter failed, mark it */
-          ptr->attention = 1;
-        }
-
-        if (fail != ptr->num_events) {
-          /* at least one counter succeeded, insert the thread in list */
-          list->num_tids++;
-          num_tids++;
-        }
-#endif
         list->num_tids++;  /* insert in any case */
         num_tids++;
       }
       closedir(thr_dir);
     }
   }
+
+  /* handle inactive processes */
+  for(i=0; i < num_inactive; i++) {
+    start_counters(inactive[i], screen, &events, options);
+  }
+  free(inactive);
+
   closedir(pid_dir);
 }
 
@@ -581,12 +652,12 @@ void accumulate_stats(const struct process_list* const list)
       /* accumulate in owner process */
       owner->cpu_percent += p->cpu_percent;
       for(zz = 0; zz < p->num_events; zz++) {
-        /* as soon as one thread has invalid value, skip entire process. */
-        if (p->values[zz] == 0xffffffff) {
+        /* as soon as one thread has an invalid value, mark the owner
+           as invalid as well (cannot compute a correct value) */
+        if (p->values[zz] == 0xffffffff)
           owner->values[zz] = 0xffffffff;
-          break;
-        }
-        owner->values[zz] += p->values[zz];
+        else
+          owner->values[zz] += p->values[zz];
       }
     }
   }
